@@ -1,8 +1,9 @@
-'''Services module for making modifications to database.'''
+''' Write operations / queries to easytrack's `core` and `data` tables. '''
 
 # stdlib
 from datetime import datetime
 from typing import Dict, List, Optional, Union
+import pytz
 
 # app
 from . import selectors as slc
@@ -11,6 +12,8 @@ from . import wrappers
 from .utils import notnull
 from .settings import ColumnTypes
 
+# region user
+
 
 def create_user(
     email: str,
@@ -47,64 +50,12 @@ def set_user_session_key(
     user.save()
 
 
-def add_campaign_participant(
-    campaign: mdl.Campaign,
-    add_user: mdl.User,
-) -> bool:
-    """
-    Binds user with campaign, making a participant.
-    :param add_user: User object to be bound to a campaign
-    :param campaign: Campaign object that user binds with
-    :return: whether user has been bound (false if already bound)
-    """
+# endregion
 
-    if slc.is_participant(user = notnull(add_user), campaign = notnull(campaign)):
-        return False
-
-    # 1. bind the user to campaign
-    participant = mdl.Participant.create(campaign = campaign, user = add_user)
-
-    # 2. create a new data table for the participant
-    for data_source in slc.get_campaign_data_sources(campaign = campaign):
-        wrappers.DataTable(participant = participant, data_source = data_source).create_table()
-        wrappers.AggDataTable(participant = participant, data_source = data_source).create_table()
-
-    return True
+# region campaign
 
 
-def add_supervisor_to_campaign(
-    supervisor: mdl.Supervisor,
-    new_user: mdl.User,
-) -> bool:
-    """
-    Binds user with campaign, making a supervisor.
-    :param new_user: User object to be bound to a campaign
-    :param supervisor: Supervisor object that user binds with
-    :return: whether user has been bound (false if already bound)
-    """
-
-    campaign: mdl.Campaign = notnull(supervisor).campaign
-
-    if slc.is_supervisor(user = notnull(new_user), campaign = notnull(campaign)):
-        return False
-
-    mdl.Supervisor.create(campaign = campaign, user = new_user)
-    return True
-
-
-def remove_supervisor_from_campaign(old_supervisor: mdl.Supervisor):
-    """
-    Removes a supervisor from a campaign
-    :param old_supervisor: Supervisor object to be removed
-    :return: None
-    """
-
-    campaign: mdl.Campaign = notnull(old_supervisor).campaign
-    if old_supervisor.user != campaign.owner:
-        old_supervisor.delete()
-
-
-def create_campaign(   # pylint: disable=too-many-arguments
+def create_campaign(
     owner: mdl.User,
     name: str,
     description: Optional[str],
@@ -121,6 +72,7 @@ def create_campaign(   # pylint: disable=too-many-arguments
     :param data_sources: list of data sources to be added to the campaign
     :return: Campaign object
     """
+    # pylint: disable=too-many-arguments
 
     # 0. validate the arguments
     today = datetime.today().replace(hour = 0, minute = 0, second = 0, microsecond = 0)
@@ -151,48 +103,6 @@ def create_campaign(   # pylint: disable=too-many-arguments
     return campaign
 
 
-def add_campaign_data_source(
-    campaign: mdl.Campaign,
-    data_source: mdl.DataSource,
-) -> bool:
-    """
-    Adds a data source to campaign
-    :param campaign: the campaign to add data source to
-    :param data_source: data source being added
-    :return: whether data source has been added (false if already added)
-    """
-
-    if slc.is_campaign_data_source(campaign = campaign, data_source = data_source):
-        return False
-
-    mdl.CampaignDataSource.create(campaign = campaign, data_source = data_source)
-    for participant in slc.get_campaign_participants(campaign):
-        wrappers.DataTable(participant, data_source).create_table()
-    return True
-
-
-def remove_campaign_data_source(
-    campaign: mdl.Campaign,
-    data_source: mdl.DataSource,
-):
-    """
-    Removes a data source from a campaign
-    :param campaign: the campaign to remove data source from
-    :param data_source: data source being removed
-    :return: None
-    """
-
-    if not slc.is_campaign_data_source(campaign = campaign, data_source = data_source):
-        return
-
-    campaign_data_sources = mdl.CampaignDataSource.filter(
-        campaign = campaign,
-        data_source = data_source,
-    )
-    for campaign_data_source in campaign_data_sources:
-        campaign_data_source.delete_instance()
-
-
 def update_campaign(
     supervisor: mdl.Supervisor,
     name: str,
@@ -238,6 +148,169 @@ def delete_campaign(supervisor: mdl.Supervisor):
         campaign.delete_instance()
 
 
+# endregion
+
+# region participant
+
+
+def add_campaign_participant(
+    campaign: mdl.Campaign,
+    add_user: mdl.User,
+) -> bool:
+    """
+    Binds user with campaign, making a participant.
+    :param add_user: User object to be bound to a campaign
+    :param campaign: Campaign object that user binds with
+    :return: whether user has been bound (false if already bound)
+    """
+
+    if slc.is_participant(user = notnull(add_user), campaign = notnull(campaign)):
+        return False
+
+    # 1. bind the user to campaign
+    participant = mdl.Participant.create(campaign = campaign, user = add_user)
+
+    # 2. create a new data table for the participant
+    for data_source in slc.get_campaign_data_sources(campaign = campaign):
+        wrappers.DataTable(participant = participant, data_source = data_source).create_table()
+        wrappers.AggDataTable(participant = participant, data_source = data_source).create_table()
+
+    return True
+
+
+# endregion
+
+# region supervisor
+
+
+def add_supervisor_to_campaign(
+    supervisor: mdl.Supervisor,
+    new_user: mdl.User,
+) -> bool:
+    """
+    Binds user with campaign, making a supervisor.
+    :param new_user: User object to be bound to a campaign
+    :param supervisor: Supervisor object that user binds with
+    :return: whether user has been bound (false if already bound)
+    """
+
+    campaign: mdl.Campaign = notnull(supervisor).campaign
+
+    if slc.is_supervisor(user = notnull(new_user), campaign = notnull(campaign)):
+        return False
+
+    mdl.Supervisor.create(campaign = campaign, user = new_user)
+    return True
+
+
+def remove_supervisor_from_campaign(old_supervisor: mdl.Supervisor):
+    """
+    Removes a supervisor from a campaign
+    :param old_supervisor: Supervisor object to be removed
+    :return: None
+    """
+
+    campaign: mdl.Campaign = notnull(old_supervisor).campaign
+    if old_supervisor.user != campaign.owner:
+        old_supervisor.delete()
+
+
+# endregion
+
+# region data source
+
+
+def create_data_source(
+    name: str,
+    columns: List[mdl.DataSourceColumn],
+) -> mdl.DataSource:
+    """
+    Creates a data source object in database and returns DataSource object
+    :param name: name of the data source
+    :param columns: list of columns of the data source
+    :return: DataSource object
+    """
+
+    # assert that name is not empty
+    if not name:
+        raise ValueError('Name cannot be empty!')
+
+    # assert that columns are not empty
+    if not columns:
+        raise ValueError('columns cannot be empty!')
+
+    # check if data source already exists (by name)
+    data_source = mdl.DataSource.get_or_none(name = notnull(name))
+    if data_source:
+        return data_source
+
+    # create data source
+    data_source = mdl.DataSource.create(name = name)
+
+    # add timestamp (reserved) column
+    timestamp_column = mdl.Column.create(
+        name = ColumnTypes.TIMESTAMP.name,
+        column_type = 'timestamp',
+        is_categorical = False,
+    )
+    mdl.DataSourceColumn.create(data_source = data_source, column = timestamp_column)
+
+    # add columns (except reserved `timestamp` column)
+    for column in columns:
+        if column.name == ColumnTypes.TIMESTAMP.name:
+            continue   # skip reserved `timestamp` column (already added)
+        mdl.DataSourceColumn.create(data_source = data_source, column = column)
+
+    return data_source
+
+
+def add_campaign_data_source(
+    campaign: mdl.Campaign,
+    data_source: mdl.DataSource,
+) -> bool:
+    """
+    Adds a data source to campaign
+    :param campaign: the campaign to add data source to
+    :param data_source: data source being added
+    :return: whether data source has been added (false if already added)
+    """
+
+    if slc.is_campaign_data_source(campaign = campaign, data_source = data_source):
+        return False
+
+    mdl.CampaignDataSource.create(campaign = campaign, data_source = data_source)
+    for participant in slc.get_campaign_participants(campaign):
+        wrappers.DataTable(participant, data_source).create_table()
+    return True
+
+
+def remove_campaign_data_source(
+    campaign: mdl.Campaign,
+    data_source: mdl.DataSource,
+):
+    """
+    Removes a data source from a campaign
+    :param campaign: the campaign to remove data source from
+    :param data_source: data source being removed
+    :return: None
+    """
+
+    if not slc.is_campaign_data_source(campaign = campaign, data_source = data_source):
+        return
+
+    campaign_data_sources = mdl.CampaignDataSource.filter(
+        campaign = campaign,
+        data_source = data_source,
+    )
+    for campaign_data_source in campaign_data_sources:
+        campaign_data_source.delete_instance()
+
+
+# endregion
+
+# region column
+
+
 def create_column(
     name: str,
     column_type: str,
@@ -310,48 +383,56 @@ def create_column(
     )
 
 
-def create_data_source(
-    name: str,
-    columns: List[mdl.DataSourceColumn],
-) -> mdl.DataSource:
+# endregion
+
+# region hourly stats
+
+
+def create_hourly_stats(
+    participant: mdl.Participant,
+    data_source: mdl.DataSource,
+    hour_timestamp: datetime,
+    amount: Dict[int, int],
+):
     """
-    Creates a data source object in database and returns DataSource object
-    :param name: name of the data source
-    :param columns: list of columns of the data source
-    :return: DataSource object
+    Verifies column ids in `amount` and creates hourly stats at a given hour `hour_timestamp`
+    for particular participant and data source. Note that the `hour_timestamp` is rounded down
+    to the nearest hour.
+    :param participant: participant of a campaign
+    :param data_source: data source of the data record
+    :param hour_timestamp: timestamp of the hour
+    :param amounts: dict of amounts (key: column id, value: amount of data records)
     """
 
-    # assert that name is not empty
-    if not name:
-        raise ValueError('Name cannot be empty!')
+    # preprocess timestamp (i.e. round down to nearest hour)
+    # (1) verify that timestamp is a datetime instance
+    if not isinstance(hour_timestamp, datetime):
+        raise ValueError('`hour_timestamp` must be a datetime object!')
+    # (2) remove timezone info (first convert to UTC, then remove timezone info)
+    hour_timestamp = hour_timestamp.astimezone(tz = pytz.utc)
+    hour_timestamp = hour_timestamp.replace(tzinfo = None)
+    # (3) round down to nearest hour
+    hour_timestamp = hour_timestamp.replace(minute = 0, second = 0, microsecond = 0)
 
-    # assert that columns are not empty
-    if not columns:
-        raise ValueError('columns cannot be empty!')
+    # verify column ids are valid
+    column_ids = {column.id for column in slc.get_data_source_columns(data_source = data_source)}
+    for column_id in amount.keys():
+        if column_id not in column_ids:
+            raise ValueError(f'Invalid column id: {column_id}')
 
-    # check if data source already exists (by name)
-    data_source = mdl.DataSource.get_or_none(name = notnull(name))
-    if data_source:
-        return data_source
+    # create hourly stats (i.e. insert into database)
+    # pylint: disable=no-value-for-parameter
+    mdl.HourlyStats.insert(
+        participant = participant,
+        data_source = data_source,
+        timestamp = hour_timestamp,
+        amount = amount,
+    ).execute()
 
-    # create data source
-    data_source = mdl.DataSource.create(name = name)
 
-    # add timestamp (reserved) column
-    timestamp_column = mdl.Column.create(
-        name = ColumnTypes.TIMESTAMP.name,
-        column_type = 'timestamp',
-        is_categorical = False,
-    )
-    mdl.DataSourceColumn.create(data_source = data_source, column = timestamp_column)
+# endregion
 
-    # add columns (except reserved `timestamp` column)
-    for column in columns:
-        if column.name == ColumnTypes.TIMESTAMP.name:
-            continue   # skip reserved `timestamp` column (already added)
-        mdl.DataSourceColumn.create(data_source = data_source, column = column)
-
-    return data_source
+# region raw data
 
 
 def create_data_record(
@@ -364,11 +445,11 @@ def create_data_record(
     Creates a data record in raw data table (e.g. sensor reading)
     :param participant: participant of a campaign
     :param data_source: data source of the data record
-    :param ts: timestamp
-    :param val: value
-    :return: None
+    :param timestamp: timestamp of the data record
+    :param value: value of the data record (dict of column id and value)
     """
 
+    # NOTE: verification is already done in wrappers.DataTable.insert() function
     wrappers.DataTable(participant = participant, data_source = data_source).insert(
         timestamp = timestamp,
         value = value,
@@ -390,6 +471,7 @@ def create_data_records(
     :return: None
     """
 
+    # NOTE: verification is already done in wrappers.DataTable.insert() function
     data_sources: Dict[int, mdl.DataSource] = {}   # dict()
     for timestamp, data_source_id, value in zip(timestamps, data_source_ids, values):
 
@@ -426,3 +508,6 @@ def dump_data(
 
     data_table = wrappers.DataTable(participant = participant, data_source = data_source)
     return data_table.dump_to_file()
+
+
+# endregion
